# flake8: noqa
"""
Random SubPackage
=================

Utilities
---------
Utilities to Support Generation of Random Arrays or Matrices
1. probvec
2. sample_without_replacement

.. Future Work
   -----------
   1. AR1 Function
"""

from .utilities import probvec, sample_without_replacement, draw
